"""
Unit tests for ionerdss.model.pdb.template_builder

Tests the TemplateBuilder class and its template generation capabilities.
"""

import unittest
from unittest.mock import Mock, patch
import numpy as np
from collections import defaultdict

from ionerdss.model.pdb.template_builder import (
    TemplateBuilder, GeometricSignature
)
from ionerdss.model.pdb.parser import PDBParser
from ionerdss.model.pdb.coarse_graining import CoarseGrainer, CoarseGrainedChain, InterfaceString
from ionerdss.model.pdb.chain_grouping import ChainGrouper, ChainGroup
from ionerdss.model.pdb.hyperparameters import PDBModelHyperparameters
from ionerdss.model.pdb.file_manager import WorkspaceManager
from ionerdss.model.components.types import MoleculeType, InterfaceType
from ionerdss.model.components.units import Units


class TestGeometricSignature(unittest.TestCase):
    """Test cases for GeometricSignature dataclass."""

    def test_geometric_signature_creation(self):
        """Test GeometricSignature initialization."""
        signature = GeometricSignature(
            d_i=5.0,
            d_j=6.0,
            theta_i=1.2,
            theta_j=1.5
        )

        self.assertEqual(signature.d_i, 5.0)
        self.assertEqual(signature.d_j, 6.0)
        self.assertEqual(signature.theta_i, 1.2)
        self.assertEqual(signature.theta_j, 1.5)

    def test_normalize(self):
        """Test signature normalization."""
        signature = GeometricSignature(
            d_i=5.123456789,
            d_j=6.987654321,
            theta_i=1.234567,
            theta_j=1.567890
        )

        normalized = signature.normalize(precision=3)
        expected = (5.123, 6.988, 1.235, 1.568)

        self.assertEqual(normalized, expected)

    def test_is_similar_to(self):
        """Test signature similarity comparison."""
        signature1 = GeometricSignature(5.0, 6.0, 1.2, 1.5)
        signature2 = GeometricSignature(5.1, 6.1, 1.25, 1.55)
        signature3 = GeometricSignature(7.0, 8.0, 2.0, 2.5)

        # Similar signatures
        self.assertTrue(signature1.is_similar_to(
            signature2, distance_threshold=0.2, angle_threshold=0.1
        ))

        # Dissimilar signatures
        self.assertFalse(signature1.is_similar_to(
            signature3, distance_threshold=0.2, angle_threshold=0.1
        ))

    def test_is_homotypic(self):
        """Test homotypic signature detection."""
        # Homotypic signature (symmetric)
        homotypic = GeometricSignature(5.0, 5.1, 1.2, 1.25)
        self.assertTrue(homotypic.is_homotypic(
            distance_threshold=0.2, angle_threshold=0.1
        ))

        # Heterotypic signature (asymmetric)
        heterotypic = GeometricSignature(5.0, 7.0, 1.2, 2.0)
        self.assertFalse(heterotypic.is_homotypic(
            distance_threshold=0.2, angle_threshold=0.1
        ))


class TestTemplateBuilder(unittest.TestCase):
    """Test cases for TemplateBuilder class."""

    def setUp(self):
        """Set up test fixtures."""
        # Create mock components
        self.parser = Mock(spec=PDBParser)
        self.coarse_grainer = Mock(spec=CoarseGrainer)
        self.chain_grouper = Mock(spec=ChainGrouper)
        self.hyperparams = Mock(spec=PDBModelHyperparameters)
        self.workspace_manager = Mock(spec=WorkspaceManager)
        self.workspace_manager.logger = Mock()

        # Set up mock data
        self._setup_mock_data()
        
        # Add chains attribute to mock coarse_grainer
        self.coarse_grainer.chains = self.coarse_grainer.get_coarse_grained_chains.return_value

    def _setup_mock_data(self):
        """Set up mock data for testing."""
        # Mock hyperparameters
        self.hyperparams.homodimer_distance_threshold = 1.0
        self.hyperparams.homodimer_angle_threshold = 0.2
        self.hyperparams.interface_type_assignment_distance_threshold = 1.0
        self.hyperparams.interface_type_assignment_angle_threshold = 0.2
        self.hyperparams.signature_precision = 6
        self.hyperparams.steric_clash_mode = "off"

        # Mock coarse-grained chains
        chain_a = Mock(spec=CoarseGrainedChain)
        chain_a.chain_id = "A"
        chain_a.com = np.array([0.0, 0.0, 0.0])
        chain_a.radius = 15.0  # Angstroms

        chain_b = Mock(spec=CoarseGrainedChain)
        chain_b.chain_id = "B"
        chain_b.com = np.array([10.0, 0.0, 0.0])
        chain_b.radius = 12.0  # Angstroms

        self.coarse_grainer.get_coarse_grained_chains.return_value = {
            "A": chain_a,
            "B": chain_b
        }

        # Mock interfaces
        interface_ab = Mock(spec=InterfaceString)
        interface_ab.chain_i = "A"
        interface_ab.chain_j = "B"
        interface_ab.coord_i = np.array([5.0, 0.0, 0.0])
        interface_ab.coord_j = np.array([5.0, 0.0, 0.0])
        interface_ab.residues_i = {1, 2, 3}
        interface_ab.residues_j = {4, 5, 6}
        interface_ab.energy = -5.0

        self.coarse_grainer.get_interfaces.return_value = [interface_ab]

        # Mock chain groups
        group_a = Mock(spec=ChainGroup)
        group_a.representative = "A"
        group_a.members = ["A"]
        group_a.grouping_method = "sequence_similarity"

        group_b = Mock(spec=ChainGroup)
        group_b.representative = "B"
        group_b.members = ["B"]
        group_b.grouping_method = "sequence_similarity"

        self.chain_grouper.get_groups.return_value = [group_a, group_b]
        self.chain_grouper.get_group_for_chain.side_effect = lambda chain_id: {
            "A": group_a,
            "B": group_b
        }.get(chain_id)

        # Mock parser methods
        self.parser.convert_coords_to_nm.side_effect = lambda coords: coords / 10.0
        self.parser.get_chain_data.return_value = {
            'ca_coords': np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
        }

    def test_initialization(self):
        """Test TemplateBuilder initialization."""
        with patch.object(TemplateBuilder, '_build_templates'):
            with patch.object(TemplateBuilder, '_regularize_templates'):
                with patch.object(TemplateBuilder, '_detect_steric_clashes'):
                    builder = TemplateBuilder(
                        parser=self.parser,
                        coarse_grainer=self.coarse_grainer,
                        chain_grouper=self.chain_grouper,
                        hyperparams=self.hyperparams,
                        workspace_manager=self.workspace_manager
                    )

        self.assertEqual(builder.parser, self.parser)
        self.assertEqual(builder.coarse_grainer, self.coarse_grainer)
        self.assertEqual(builder.chain_grouper, self.chain_grouper)
        self.assertEqual(builder.hyperparams, self.hyperparams)
        self.assertEqual(builder.workspace_manager, self.workspace_manager)
        self.assertIsInstance(builder.units, Units)

    def test_initialization_with_custom_units(self):
        """Test TemplateBuilder initialization with custom units."""
        custom_units = Mock(spec=Units)

        with patch.object(TemplateBuilder, '_build_templates'):
            with patch.object(TemplateBuilder, '_regularize_templates'):
                with patch.object(TemplateBuilder, '_detect_steric_clashes'):
                    builder = TemplateBuilder(
                        parser=self.parser,
                        coarse_grainer=self.coarse_grainer,
                        chain_grouper=self.chain_grouper,
                        hyperparams=self.hyperparams,
                        units=custom_units,
                        workspace_manager=self.workspace_manager
                    )

        self.assertEqual(builder.units, custom_units)

    def test_generate_template_name_unique(self):
        """Test template name generation for unique representative."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.used_template_names = set()

        group = Mock(spec=ChainGroup)
        group.representative = "A"
        group.members = ["A"]

        name = builder._generate_template_name(group)

        self.assertEqual(name, "A")
        self.assertIn("A", builder.used_template_names)

    def test_generate_template_name_conflict(self):
        """Test template name generation with conflicts."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.used_template_names = {"A"}  # A is already used

        group = Mock(spec=ChainGroup)
        group.representative = "A"
        group.members = ["A", "C"]  # Multiple members

        name = builder._generate_template_name(group)

        self.assertEqual(name, "A0")
        self.assertIn("A0", builder.used_template_names)

    def test_generate_template_name_numeric_suffix(self):
        """Test template name generation with numeric suffix."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.used_template_names = {"A", "A_group"}  # Both variants used

        group = Mock(spec=ChainGroup)
        group.representative = "A"
        group.members = ["A"]

        name = builder._generate_template_name(group)

        self.assertEqual(name, "A1")
        self.assertIn("A1", builder.used_template_names)

    @unittest.skip("_build_molecule_template implementation changed - needs mock updates")
    def test_build_molecule_template(self):
        """Test molecule template building."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.coarse_grainer = self.coarse_grainer
        builder.parser = self.parser
        builder.workspace_manager = self.workspace_manager
        builder.used_template_names = set()
        builder.group_to_template = {}
        builder.molecule_templates = {}

        group = Mock(spec=ChainGroup)
        group.representative = "A"
        group.members = ["A"]
        group.grouping_method = "sequence_similarity"

        builder._build_molecule_template(group)

        # Check that template was created
        self.assertIn("A", builder.molecule_templates)
        self.assertEqual(builder.group_to_template["A"], "A")

        # Check template properties
        template = builder.molecule_templates["A"]
        self.assertIsInstance(template, MoleculeType)
        self.assertEqual(template.name, "A")
        self.assertEqual(template.radius_nm, 1.5)  # 15.0 Å / 10

    def test_calculate_geometric_signature(self):
        """Test geometric signature calculation."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.coarse_grainer = self.coarse_grainer

        interface = Mock(spec=InterfaceString)
        interface.chain_i = "A"
        interface.chain_j = "B"
        interface.coord_i = np.array([5.0, 0.0, 0.0])
        interface.coord_j = np.array([5.0, 0.0, 0.0])

        signature = builder._calculate_geometric_signature(interface)

        self.assertIsInstance(signature, GeometricSignature)
        self.assertGreater(signature.d_i, 0)
        self.assertGreater(signature.d_j, 0)
        self.assertGreaterEqual(signature.theta_i, 0)
        self.assertGreaterEqual(signature.theta_j, 0)

    def test_create_homotypic_interface_template(self):
        """Test homodimer interface template creation."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.coarse_grainer = self.coarse_grainer
        builder.parser = self.parser
        builder.workspace_manager = self.workspace_manager
        builder.hyperparams = self.hyperparams  # Add this line
        builder.interface_templates = {}
        builder.interface_signatures = {}
        builder.molecule_templates = {
            "A": Mock(spec=MoleculeType, interfaces_neighbors_map={})
        }

        interface = Mock(spec=InterfaceString)
        interface.chain_i = "A"
        interface.chain_j = "A"
        interface.coord_i = np.array([5.0, 0.0, 0.0])
        interface.residues_i = {1, 2}
        interface.residues_j = {3, 4}
        interface.energy = -3.0
        interface.residue_details_i = [Mock(id=1, name="ALA"), Mock(id=2, name="GLY")]  # Add these
        interface.residue_details_j = [Mock(id=3, name="CYS"), Mock(id=4, name="VAL")]  # Add these

        signature = GeometricSignature(5.0, 5.0, 1.0, 1.0)

        interface_name = builder._create_homotypic_interface_template(
            interface, "A", signature, 1
        )

        # Check that template was created
        expected_name = "AA1"  # Changed from "A_A_1"
        self.assertEqual(interface_name, expected_name)
        self.assertIn(expected_name, builder.interface_templates)
        self.assertIn(expected_name, builder.interface_signatures)

        # Check template properties
        template = builder.interface_templates[expected_name]
        self.assertEqual(template.this_mol_type_name, "A")
        self.assertEqual(template.partner_mol_type_name, "A")
        self.assertEqual(template.interface_index, 1)

    def test_create_heterotypic_interface_templates(self):
        """Test heterotypic interface template creation."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.coarse_grainer = self.coarse_grainer
        builder.parser = self.parser
        builder.workspace_manager = self.workspace_manager
        builder.hyperparams = self.hyperparams  # Add this line
        builder.interface_templates = {}
        builder.interface_signatures = {}
        builder.molecule_templates = {
            "A": Mock(spec=MoleculeType, interfaces_neighbors_map={}),
            "B": Mock(spec=MoleculeType, interfaces_neighbors_map={})
        }

        interface = Mock(spec=InterfaceString)
        interface.chain_i = "A"
        interface.chain_j = "B"
        interface.coord_i = np.array([5.0, 0.0, 0.0])
        interface.coord_j = np.array([5.0, 0.0, 0.0])
        interface.residues_i = {1, 2}
        interface.residues_j = {3, 4}
        interface.energy = -3.0
        interface.residue_details_i = [Mock(id=1, name="ALA"), Mock(id=2, name="GLY")]  # Add these
        interface.residue_details_j = [Mock(id=3, name="CYS"), Mock(id=4, name="VAL")]  # Add these

        signature = GeometricSignature(5.0, 5.0, 1.0, 1.2)

        interface_names = builder._create_heterotypic_interface_templates(
            interface, "A", "B", signature, 1
        )

        # Check that both templates were created
        expected_names = ["AB1", "BA1"]  # Changed from "A_B_1", "B_A_1"
        self.assertEqual(len(interface_names), 2)
        self.assertCountEqual(interface_names, expected_names)

        # Check templates exist
        for name in expected_names:
            self.assertIn(name, builder.interface_templates)
            self.assertIn(name, builder.interface_signatures)

        # Check cross-references
        template_a = builder.interface_templates["AB1"]  # Changed
        template_b = builder.interface_templates["BA1"]  # Changed
        self.assertEqual(template_a.partner_interface_type, template_b)
        self.assertEqual(template_b.partner_interface_type, template_a)

    def test_find_matching_interface_type_found(self):
        """Test finding matching interface type."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.interface_templates = {}
        builder.interface_signatures = {}
        builder.interface_templates = {}
        builder.interface_signatures = {}
        builder.workspace_manager = self.workspace_manager
        builder.hyperparams = self.hyperparams

        # Create existing interface template
        existing_template = Mock(spec=InterfaceType)
        existing_template.this_mol_type_name = "A"
        existing_template.partner_mol_type_name = "B"

        existing_signature = GeometricSignature(5.0, 6.0, 1.0, 1.2)

        builder.interface_templates["A_B_1"] = existing_template
        builder.interface_signatures["A_B_1"] = existing_signature

        # Test with similar signature
        test_signature = GeometricSignature(5.1, 6.1, 1.05, 1.25)

        match = builder._find_matching_interface_type("A", "B", test_signature)

        self.assertEqual(match, "A_B_1")

    def test_find_matching_interface_type_not_found(self):
        """Test finding matching interface type when none exists."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.interface_templates = {}
        builder.interface_signatures = {}
        builder.interface_templates = {}
        builder.interface_signatures = {}
        builder.workspace_manager = self.workspace_manager
        builder.hyperparams = self.hyperparams

        test_signature = GeometricSignature(5.0, 6.0, 1.0, 1.2)

        match = builder._find_matching_interface_type("A", "B", test_signature)

        self.assertIsNone(match)

    def test_process_interface_with_signature_existing(self):
        """Test processing interface with existing signature match."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.workspace_manager = self.workspace_manager
        builder.interface_templates = {}
        builder.chain_assigned_types = defaultdict(set)

        # Mock finding existing match
        with patch.object(builder, '_find_matching_interface_type', return_value="A_B_1"):
            interface = Mock(spec=InterfaceString)
            interface.chain_i = "A"
            interface.chain_j = "B"
            interface.coord_i = np.array([5.0, 0.0, 0.0])
            interface.coord_j = np.array([5.0, 0.0, 0.0])
            signature = GeometricSignature(5.0, 6.0, 1.0, 1.2)

            builder._process_interface_with_signature(
                interface, "A", "B", signature
            )

            # Check that interface was assigned existing type
            self.assertEqual(interface.interface_type, "A_B_1")

    def test_process_interface_with_signature_new(self):
        """Test processing interface with new signature."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.workspace_manager = self.workspace_manager
        builder.interface_templates = {}
        builder.chain_assigned_types = defaultdict(set)

        # Mock no existing match and new type creation
        with patch.object(builder, '_find_matching_interface_type', return_value=None):
            with patch.object(builder, '_create_new_interface_type', return_value=["A_B_1"]):
                interface = Mock(spec=InterfaceString)
                interface.chain_i = "A"
                interface.chain_j = "B"
                interface.coord_i = np.array([5.0, 0.0, 0.0])
                interface.coord_j = np.array([5.0, 0.0, 0.0])
                signature = GeometricSignature(5.0, 6.0, 1.0, 1.2)

                builder._process_interface_with_signature(
                    interface, "A", "B", signature
                )

                # Check that interface was assigned new type
                self.assertEqual(interface.interface_type, "A_B_1")

    def test_assign_interface_to_heterotypic_type(self):
        """Test assigning interface to heterotypic type."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.interface_templates = {}

        # Create mock interface templates
        template1 = Mock(spec=InterfaceType)
        template1.this_mol_type_name = "A"
        template1.signature = {"this_side": "i"}

        template2 = Mock(spec=InterfaceType)
        template2.this_mol_type_name = "B"
        template2.signature = {"this_side": "j"}

        builder.interface_templates["A_B_1"] = template1
        builder.interface_templates["B_A_1"] = template2

        interface = Mock(spec=InterfaceString)
        interface_type_names = ["A_B_1", "B_A_1"]

        assigned_type = builder._assign_interface_to_heterotypic_type(
            interface, interface_type_names, "A", "B"
        )

        self.assertEqual(assigned_type, "A_B_1")

    def test_get_interface_type_for_interface(self):
        """Test getting interface type for interface."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.interface_to_type_mapping = {
            "A_B_5.000_0.000_0.000": "A_B_1"
        }

        interface = Mock(spec=InterfaceString)
        interface.chain_i = "A"
        interface.chain_j = "B"
        interface.coord_i = np.array([5.0, 0.0, 0.0])

        interface_type = builder.get_interface_type_for_interface(interface)

        self.assertEqual(interface_type, "A_B_1")

    def test_get_interface_type_for_interface_not_found(self):
        """Test getting interface type for interface when not found."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.interface_to_type_mapping = {}

        interface = Mock(spec=InterfaceString)
        interface.chain_i = "A"
        interface.chain_j = "B"
        interface.coord_i = np.array([5.0, 0.0, 0.0])

        interface_type = builder.get_interface_type_for_interface(interface)

        self.assertIsNone(interface_type)

    def test_compute_rigid_transform(self):
        """Test rigid transform computation."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.parser = self.parser

        ref_data = Mock(spec=CoarseGrainedChain)
        ref_data.chain_id = "A"

        mem_data = Mock(spec=CoarseGrainedChain)
        mem_data.chain_id = "B"

        # Mock parser to return same coordinates (identity transform expected)
        self.parser.get_chain_data.return_value = {
            'ca_coords': np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        }

        transform = builder._compute_rigid_transform(ref_data, mem_data)

        # Should return 4x4 transformation matrix
        self.assertEqual(transform.shape, (4, 4))
        # For identical coordinates, should be close to identity
        np.testing.assert_allclose(transform, np.eye(4), atol=1e-10)

    @unittest.skip("regularize_group method signature changed - needs test update")
    def test_regularize_group(self):
        """Test group regularization."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.coarse_grainer = self.coarse_grainer
        builder.parser = self.parser

        # Create group with multiple members
        group = Mock(spec=ChainGroup)
        group.representative = "A"
        group.members = ["A", "C"]  # A is reference, C needs regularization

        # Add chain C to mock data
        chain_c = Mock(spec=CoarseGrainedChain)
        chain_c.chain_id = "C"
        chain_c.com = np.array([20.0, 0.0, 0.0])
        self.coarse_grainer.get_coarse_grained_chains.return_value["C"] = chain_c

        # Mock compute_rigid_transform
        with patch.object(builder, '_compute_rigid_transform', return_value=np.eye(4)):
            builder.regularize_group(group)  # Changed from _regularize_group

        # Check that transform was computed and stored
        self.assertTrue(hasattr(chain_c, 'transform_from_reference'))

    def test_detect_steric_clashes(self):
        """Test steric clash detection."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.interface_templates = {}

        # Create mock interface templates on same molecule
        template1 = Mock(spec=InterfaceType)
        template1.this_mol_type = Mock(name="A")
        template1.local_coord = np.array([0.0, 0.0, 0.0])
        template1.required_free = []

        template2 = Mock(spec=InterfaceType)
        template2.this_mol_type = template1.this_mol_type  # Same molecule
        template2.local_coord = np.array([0.2, 0.0, 0.0])  # Close position
        template2.required_free = []

        # TO BE UDPATED WHEN STERIC CLASHES ARE FIXED
        #builder.interface_templates["intf1"] = template1
        #builder.interface_templates["intf2"] = template2

        #builder._detect_steric_clashes()

        # Check that mutual exclusion was set up
        #self.assertIn("intf2", template1.required_free)
        #self.assertIn("intf1", template2.required_free)

    def test_get_molecule_templates(self):
        """Test getting molecule templates."""
        builder = TemplateBuilder.__new__(TemplateBuilder)

        template_a = Mock(spec=MoleculeType)
        template_b = Mock(spec=MoleculeType)

        builder.molecule_templates = {
            "A": template_a,
            "B": template_b
        }

        templates = builder.get_molecule_templates()

        self.assertEqual(len(templates), 2)
        self.assertIn("A", templates)
        self.assertIn("B", templates)
        self.assertEqual(templates["A"], template_a)
        self.assertEqual(templates["B"], template_b)

    def test_get_interface_templates(self):
        """Test getting interface templates."""
        builder = TemplateBuilder.__new__(TemplateBuilder)

        template_ab = Mock(spec=InterfaceType)
        template_ba = Mock(spec=InterfaceType)

        builder.interface_templates = {
            "A_B_1": template_ab,
            "B_A_1": template_ba
        }

        templates = builder.get_interface_templates()

        self.assertEqual(len(templates), 2)
        self.assertIn("A_B_1", templates)
        self.assertIn("B_A_1", templates)

    def test_get_template_name_for_group(self):
        """Test getting template name for group."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.group_to_template = {
            "A": "ProteinA",
            "B": "ProteinB"
        }

        name = builder.get_template_name_for_group("A")
        self.assertEqual(name, "ProteinA")

        name = builder.get_template_name_for_group("C")
        self.assertIsNone(name)

    def test_get_chain_name_mapping(self):
        """Test getting chain name mapping."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.chain_grouper = self.chain_grouper
        builder.group_to_template = {
            "A": "ProteinA",
            "B": "ProteinB"
        }

        mapping = builder.get_chain_name_mapping()

        expected = {
            "A": "ProteinA",
            "B": "ProteinB"
        }

        self.assertEqual(mapping, expected)

    def test_get_summary(self):
        """Test getting summary."""
        builder = TemplateBuilder.__new__(TemplateBuilder)
        builder.molecule_templates = {
            "A": Mock(spec=MoleculeType),
            "B": Mock(spec=MoleculeType)
        }

        # Mock interface templates
        template_ab = Mock(spec=InterfaceType)
        template_ab.this_mol_type_name = "A"
        template_ab.partner_mol_type_name = "B"

        template_ba = Mock(spec=InterfaceType)
        template_ba.this_mol_type_name = "B"
        template_ba.partner_mol_type_name = "A"

        builder.interface_templates = {
            "A_B_1": template_ab,
            "B_A_1": template_ba
        }

        builder.group_to_template = {"A": "A", "B": "B"}

        # Mock chain grouper for get_chain_name_mapping
        with patch.object(builder, 'get_chain_name_mapping', return_value={"A": "A", "B": "B"}):
            summary = builder.get_summary()

        self.assertEqual(summary["num_molecule_templates"], 2)
        self.assertEqual(summary["num_interface_templates"], 2)
        self.assertIn("A", summary["molecule_templates"])
        self.assertIn("B", summary["molecule_templates"])
        self.assertIn("A_B_1", summary["interface_templates"])
        self.assertIn("B_A_1", summary["interface_templates"])


class TestTemplateBuilderIntegration(unittest.TestCase):
    """Integration tests for TemplateBuilder."""

    def setUp(self):
        """Set up integration test fixtures."""
        self.workspace_manager = Mock(spec=WorkspaceManager)
        self.workspace_manager.logger = Mock()

    def test_complete_template_building_workflow(self):
        """Test complete template building workflow."""
        # Create comprehensive mock setup
        parser = Mock(spec=PDBParser)
        coarse_grainer = Mock(spec=CoarseGrainer)
        chain_grouper = Mock(spec=ChainGrouper)
        hyperparams = Mock(spec=PDBModelHyperparameters)

        # Set up realistic mock data
        self._setup_integration_mocks(
            parser, coarse_grainer, chain_grouper, hyperparams)

        # Build templates
        builder = TemplateBuilder(
            parser=parser,
            coarse_grainer=coarse_grainer,
            chain_grouper=chain_grouper,
            hyperparams=hyperparams,
            workspace_manager=self.workspace_manager
        )

        # Verify templates were created
        molecule_templates = builder.get_molecule_templates()
        interface_templates = builder.get_interface_templates()

        self.assertGreater(len(molecule_templates), 0)
        self.assertGreater(len(interface_templates), 0)

        # Verify summary
        summary = builder.get_summary()
        self.assertGreater(summary["num_molecule_templates"], 0)
        self.assertGreater(summary["num_interface_templates"], 0)

    def _setup_integration_mocks(self, parser, coarse_grainer, chain_grouper, hyperparams):
        """Set up comprehensive mocks for integration testing."""
        # Mock hyperparameters
        hyperparams.homodimer_distance_threshold = 1.0
        hyperparams.homodimer_angle_threshold = 0.2
        hyperparams.signature_precision = 6
        hyperparams.signature_precision = 6
        hyperparams.steric_clash_mode = "off"
        hyperparams.template_regularization_strength = 0.5
        hyperparams.min_chain_length = 4

        # Mock coarse-grained chains
        chain_a = Mock(spec=CoarseGrainedChain)
        chain_a.chain_id = "A"
        chain_a.com = np.array([0.0, 0.0, 0.0])
        chain_a.radius = 15.0

        coarse_grainer.get_coarse_grained_chains.return_value = {"A": chain_a}
        coarse_grainer.chains = {"A": chain_a}

        # Mock interfaces (self-interaction)
        interface = Mock(spec=InterfaceString)
        interface.chain_i = "A"
        interface.chain_j = "A"
        interface.coord_i = np.array([5.0, 0.0, 0.0])
        interface.coord_j = np.array([-5.0, 0.0, 0.0])
        interface.residues_i = {1, 2}
        interface.residues_j = {3, 4}
        interface.energy = -3.0
        interface.residue_details_i = [Mock(id=1, name="ALA"), Mock(id=2, name="GLY")]
        interface.residue_details_j = [Mock(id=3, name="CYS"), Mock(id=4, name="VAL")]
        interface.get_residue_sequence_i.return_value = "AG"
        interface.get_residue_sequence_j.return_value = "CV"
        interface.get_residue_composition_i.return_value = {"ALA": 1, "GLY": 1}
        interface.get_residue_composition_j.return_value = {"CYS": 1, "VAL": 1}

        coarse_grainer.get_interfaces.return_value = [interface]
        coarse_grainer.interfaces = [interface]

        # Mock chain group
        group = Mock(spec=ChainGroup)
        group.representative = "A"
        group.members = ["A"]
        group.grouping_method = "sequence_similarity"

        chain_grouper.get_groups.return_value = [group]
        chain_grouper.get_group_for_chain.return_value = group

        # Mock parser
        parser.convert_coords_to_nm.side_effect = lambda coords: coords / 10.0
        parser.get_chain_data.return_value = {
            'ca_coords': np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
        }


if __name__ == '__main__':
    # Run with verbose output
    unittest.main(verbosity=2)



"""
capsid_sphere_pipeline.py

Modularized driver for spherical reconstruction pipeline.
Generalized from HIV Gag to support arbitrary structure mapped to sphere.

TODO:
(1) Check for distance between center of masses and regularize
(2) Check if the subunits are rotationally aligned and rotate them to the correct position if not
(3) Figure out curvature from center of mass and calculate the radius of fitted sphere
    - RMSD based gradient descent fitting?

"""
import os
import numpy as np
import pandas as pd
from .geometry import calculate_gradient, calculate_rmsd


def run_spherical_reconstruction_pipeline(pdb_file, save_dir, options=None):
    """
    Reconstructs and outputs a spherical structure by projecting subunits onto a defined sphere.

    Parameters
    ----------
    pdb_file : str
        Path to the input file containing center and interface positions per subunit.
    save_dir : str
        Directory to save output files.
    options : dict or None
        Optional parameters:
            - 'num_units': int, number of repeating units (default=18)
            - 'points_per_unit': int, number of points per unit (default=6)
            - 'radius': float, target radius for projection (nm)
            - 'fit_sphere': bool, whether to compute best-fit sphere (default=True)
    """
    # Parse input options
    options = options or {}
    num_units = options.get("num_units", 18)
    points_per_unit = options.get("points_per_unit", 6)
    radius = options.get("radius", 25.0)
    fit_sphere = options.get("fit_sphere", True)

    # Step 1: Load and normalize positions
    df = pd.read_csv(pdb_file, header=None, sep=r'\s+')
    df.columns = ['x', 'y', 'z']
    df = df / 10.0  # Convert from Å to nm
    positions_vec = df.to_numpy()

    # Step 2: Determine or fit sphere
    center_vec = np.array([positions_vec[points_per_unit * i] for i in range(num_units)])

    if fit_sphere:
        xyz_r = [0, 0, 0, 70]
        force = calculate_gradient(center_vec, xyz_r)
        rmsd_old = calculate_rmsd(center_vec, xyz_r)
        while np.linalg.norm(force) > 0.01:
            step = 1.0
            while True:
                trial = np.array(xyz_r) - step * force
                rmsd_new = calculate_rmsd(center_vec, trial)
                if rmsd_new < rmsd_old:
                    break
                step *= 0.8
            xyz_r = trial
            rmsd_old = rmsd_new
            force = calculate_gradient(center_vec, xyz_r)
        x0, y0, z0, r0 = xyz_r
        print("Fitted sphere center and radius:", xyz_r)
    else:
        x0, y0, z0 = options.get("center", [0.0, 0.0, 0.0])
        r0 = radius

    # Step 3: Normalize positions to origin + surface
    shift_vec = np.array([x0, y0, z0])
    positions_vec -= shift_vec
    center_vec -= shift_vec
    for i in range(num_units):
        shift = (center_vec[i] / np.linalg.norm(center_vec[i]) * r0) - center_vec[i]
        center_vec[i] += shift
        for j in range(points_per_unit):
            positions_vec[points_per_unit * i + j] += shift

    # Step 4: Output the constructed positions
    os.makedirs(save_dir, exist_ok=True)
    coord_out = os.path.join(save_dir, f"coordR{int(radius)}.txt")
    with open(coord_out, 'w', encoding="utf-8") as f:
        for i in range(num_units):
            unit = positions_vec[points_per_unit * i : points_per_unit * (i + 1)]
            f.write(f"UNIT{i:02d}\n")
            for xyz in unit:
                f.write(f"{xyz[0]:.8f} {xyz[1]:.8f} {xyz[2]:.8f}\n")

    print("Spherical reconstruction coordinates written to:", coord_out)

    return



"""
ionerdss.model.components.system

System-level container for complete molecular simulation models.

This module provides the System class which serves as the top-level container
for all molecular simulation components including molecule types, instances,
interface definitions, and reaction rules. The System manages the complete
state of a molecular model and provides serialization capabilities for
persistence and data exchange.

Classes:
    System: Complete molecular simulation system containing all registries,
        configuration data, and serialization methods.

Key Features:
    - Centralized management of all molecular components
    - Complete serialization/deserialization to/from JSON
    - Workspace path management for file organization
    - PDB structure integration support
    - Unit system management
    - Registry-based component organization

Architecture Overview:
    The System class acts as the root container that holds:
    - MoleculeTypeRegistry: Templates for all molecule types
    - MoleculeInstanceRegistry: Runtime instances of molecules
    - InterfaceTypeRegistry: Templates for all interface types  
    - InterfaceInstanceRegistry: Runtime instances of interfaces
    - Configuration data: workspace path, PDB ID, units
    
    All data can be serialized to JSON and fully reconstructed, maintaining
    all relationships and references between components.

Example Usage:
    ```python
    # Create a new system
    system = System(
        workspace_path="/path/to/workspace",
        pdb_id="1ABC",
        units=Units()
    )
    
    # Add components to registries
    system.molecule_types.add(protein_a)
    system.interface_types.add(binding_site)
    
    # Save to file
    system.to_json("model.json")
    
    # Load from file
    restored_system = System.from_json("model.json")
    ```

Dependencies:
    - json: For serialization/deserialization
    - pathlib: For path handling
    - All component modules: types, instances, registry, units
"""
import json
import numpy as np
from pathlib import Path
from typing import Dict, List, Optional, Any, Union

from ionerdss.model.components.units import Units
from ionerdss.model.components.registry import (
    MoleculeTypeRegistry, MoleculeInstanceRegistry,
    InterfaceTypeRegistry, InterfaceInstanceRegistry
)


class System:
    """Complete molecular system containing all components and registries.

    Attributes:
        workspace_path: Path to workspace directory.
        pdb_id: PDB identifier for this system.
        units: Unit system used throughout the system.
        molecule_types: Registry of molecule type definitions.
        molecule_instances: Registry of molecule instances.
        interface_types: Registry of interface type definitions.
        interface_instances: Registry of interface instances.
    """

    def __init__(self, workspace_path: str, pdb_id: Optional[str] = None,
                 units: Optional[Units] = None):
        """Initialize system with empty registries.

        Args:
            workspace_path: Path to workspace directory.
            pdb_id: PDB identifier (optional).
            units: Unit system (defaults to standard units).
        """
        self.workspace_path = workspace_path
        self.pdb_id = pdb_id
        self.units = units or Units()

        # Initialize registries
        self.molecule_types = MoleculeTypeRegistry()
        self.molecule_instances = MoleculeInstanceRegistry()
        self.interface_types = InterfaceTypeRegistry()
        self.interface_instances = InterfaceInstanceRegistry()

    def _convert_numpy_types(self, obj: Any) -> Any:
        """Recursively convert NumPy types to native Python types for JSON serialization.

        Args:
            obj: Object that may contain NumPy types.

        Returns:
            Object with NumPy types converted to native Python types.
        """
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, dict):
            return {key: self._convert_numpy_types(value) for key, value in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [self._convert_numpy_types(item) for item in obj]
        else:
            return obj

    def to_dict(self) -> Dict[str, Any]:
        """Convert system to dictionary representation.

        Returns:
            Dictionary containing complete system data.
        """
        system_data = {
            "metadata": {
                "workspace_path": self.workspace_path,
                "pdb_id": self.pdb_id,
                "units": self.units.to_dict()
            },
            "registries": {
                "molecule_types": self.molecule_types.to_list(),
                "molecule_instances": self.molecule_instances.to_list(),
                "interface_types": self.interface_types.to_list(),
                "interface_instances": self.interface_instances.to_list()
            }
        }

        # Convert NumPy types to native Python types
        return self._convert_numpy_types(system_data)

    def to_json(self, filepath: Union[str, Path]) -> None:
        """Save system to JSON file.

        Args:
            filepath: Path to output JSON file.
        """
        filepath = Path(filepath)

        # Get system data with NumPy types converted
        system_data = self.to_dict()

        class SystemJSONEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                elif hasattr(obj, 'to_dict'):
                    return obj.to_dict()
                elif isinstance(obj, set):
                    return list(obj)
                else:
                    return super().default(obj)
        
        system_data = self.to_dict()
        
        with open(filepath, 'w', encoding='utf-8') as f:
            json.dump(system_data, f, indent=2, ensure_ascii=False, cls=SystemJSONEncoder)


    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "System":
        """Create system from dictionary representation.

        Args:
            data: Dictionary containing system data.

        Returns:
            New System instance.
        """
        # Extract metadata
        metadata = data.get("metadata", {})
        workspace_path = metadata.get("workspace_path", ".")
        pdb_id = metadata.get("pdb_id")
        units = Units.from_dict(metadata.get("units", {}))

        # Create system
        system = cls(workspace_path=workspace_path, pdb_id=pdb_id, units=units)

        # Load registries
        registries = data.get("registries", {})

        system.molecule_types = MoleculeTypeRegistry.from_list(
            registries.get("molecule_types", [])
        )
        system.molecule_instances = MoleculeInstanceRegistry.from_list(
            registries.get("molecule_instances", [])
        )
        system.interface_types = InterfaceTypeRegistry.from_list(
            registries.get("interface_types", [])
        )
        system.interface_instances = InterfaceInstanceRegistry.from_list(
            registries.get("interface_instances", [])
        )

        # Rebuild cross-references
        system._rebuild_cross_references()

        return system

    @classmethod
    def from_json(cls, filepath: Union[str, Path]) -> "System":
        """Load system from JSON file.

        Args:
            filepath: Path to JSON file.

        Returns:
            New System instance.
        """
        filepath = Path(filepath)

        with open(filepath, 'r', encoding='utf-8') as f:
            data = json.load(f)

        return cls.from_dict(data)

    def _rebuild_cross_references(self) -> None:
        """Rebuild cross-references between components after loading.

        Internal method that restores object references between components
        that were lost during JSON serialization. This includes linking:
        - InterfaceTypes to their parent MoleculeTypes
        - InterfaceTypes to their partner InterfaceTypes
        - MoleculeInstances to their MoleculeTypes
        - InterfaceInstances to their InterfaceTypes and MoleculeInstances

        This method is called automatically by from_json() and should not
        be called directly by user code.
        """
        # Rebuild molecule type references in interface types
        for interface_type in self.interface_types:
            # Link to parent molecule type
            if interface_type.this_mol_type_name in self.molecule_types:
                interface_type.this_mol_type = self.molecule_types.get(
                    interface_type.this_mol_type_name
                )

            # Link to partner molecule type
            if interface_type.partner_mol_type_name in self.molecule_types:
                interface_type.partner_mol_type = self.molecule_types.get(
                    interface_type.partner_mol_type_name
                )

        # Rebuild partner interface references
        for interface_type in self.interface_types:
            partner_name = f"{interface_type.partner_mol_type_name}_{interface_type.this_mol_type_name}_{interface_type.interface_index}"
            if partner_name in self.interface_types:
                interface_type.partner_interface_type = self.interface_types.get(
                    partner_name)

        # Rebuild molecule instance references to their types
        for mol_instance in self.molecule_instances:
            if hasattr(mol_instance, 'molecule_type') and mol_instance.molecule_type:
                type_name = mol_instance.molecule_type.name if hasattr(
                    mol_instance.molecule_type, 'name') else str(mol_instance.molecule_type)
                if type_name in self.molecule_types:
                    mol_instance.molecule_type = self.molecule_types.get(
                        type_name)

        # Rebuild interface instance references
        for interface_instance in self.interface_instances:
            # Link to interface type
            interface_name = interface_instance.get_name()
            if interface_name in self.interface_types:
                interface_instance.interface_type = self.interface_types.get(
                    interface_name)

            # Link to parent molecule instance
            if hasattr(interface_instance, 'this_mol_name'):
                if interface_instance.this_mol_name in self.molecule_instances:
                    interface_instance.this_mol = self.molecule_instances.get(
                        interface_instance.this_mol_name
                    )

    def validate_system(self) -> Dict[str, List[str]]:
        """Validate system integrity.

        Returns:
            Dictionary with 'errors' and 'warnings' lists.
        """
        errors = []
        warnings = []

        # Check that we have some components
        if len(self.molecule_types.molecule_types) == 0:
            errors.append("No molecule types defined")

        if len(self.molecule_instances.molecule_instances) == 0:
            warnings.append("No molecule instances defined")

        # Check interface consistency
        for interface_type in self.interface_types.interface_types.values():
            # Check that referenced molecule types exist
            if interface_type.this_mol_type_name not in self.molecule_types.molecule_types:
                errors.append(
                    f"Interface type {interface_type.get_name()} references unknown molecule type: {interface_type.this_mol_type_name}")

            if interface_type.partner_mol_type_name not in self.molecule_types.molecule_types:
                errors.append(
                    f"Interface type {interface_type.get_name()} references unknown partner molecule type: {interface_type.partner_mol_type_name}")

        # Check instance consistency
        for mol_instance in self.molecule_instances.molecule_instances.values():
            if mol_instance.molecule_type and mol_instance.molecule_type.name not in self.molecule_types.molecule_types:
                errors.append(
                    f"Molecule instance {mol_instance.name} references unknown molecule type: {mol_instance.molecule_type.name}")

        return {
            "errors": errors,
            "warnings": warnings
        }

    def get_summary(self) -> Dict[str, Any]:
        """Get summary statistics of the system.

        Returns:
            Dictionary with system statistics.
        """
        return {
            "workspace_path": self.workspace_path,
            "pdb_id": self.pdb_id,
            "units": self.units.to_dict(),
            "molecule_types_count": len(self.molecule_types.molecule_types),
            "molecule_instances_count": len(self.molecule_instances.molecule_instances),
            "interface_types_count": len(self.interface_types.interface_types),
            "interface_instances_count": len(self.interface_instances.interface_instances),
            "molecule_type_names": list(self.molecule_types.molecule_types.keys()),
            "interface_type_names": list(self.interface_types.interface_types.keys())
        }

    def __str__(self) -> str:
        """String representation of the system."""
        return (f"System(pdb_id={self.pdb_id}, "
                f"molecule_types={len(self.molecule_types.molecule_types)}, "
                f"molecule_instances={len(self.molecule_instances.molecule_instances)}, "
                f"interface_types={len(self.interface_types.interface_types)}, "
                f"interface_instances={len(self.interface_instances.interface_instances)})")

    def __repr__(self) -> str:
        """Detailed representation of the system."""
        return self.__str__()
